import json
import logging
from pathlib import Path
from pprint import pprint

from xmlschema import *
from xmlschema.validators.complex_types import XsdComplexType
from xmlschema.validators.simple_types import XsdAtomicRestriction, XsdAtomicBuiltin
from xmlschema.validators.groups import *

schemas = [ "103120/schema/ts_103120_Authorisation.xsd",
            "103120/schema/ts_103120_Common.xsd",
            "103120/schema/ts_103120_Delivery.xsd",
            "103120/schema/ts_103120_Document.xsd",
            "103120/schema/ts_103120_Notification.xsd",
            "103120/schema/ts_103120_Task.xsd",
            "103280/TS_103_280.xsd",
            "testing/deps/xmldsig/xmldsig-core-schema.xsd"]

schemaLocations = []
for schemaFile in schemas:
    try:
        xs = XMLSchema(schemaFile, validation='skip')
        schemaLocations.append((xs.target_namespace, str(Path(schemaFile).resolve())))
        print (" [ {0}  ->  {1} ]".format(xs.target_namespace, schemaFile))
    except XMLSchemaParseError as ex:
        print (" [ {0} failed to parse:  {1} ]".format(schemaFile, ex))

coreSchema = XMLSchema('103120/schema/ts_103120_Core.xsd', locations = schemaLocations)

js = {
    "$id" : "core",
    "oneOf" : [],
    "$defs" : {}
}

ns_mappings = {
    'http://uri.etsi.org/03120/common/2019/10/Core' : 'core',
    'http://uri.etsi.org/03120/common/2016/02/Common' : 'common',
    'http://uri.etsi.org/03280/common/2017/07' : 'etsi103280',
    'http://www.w3.org/2000/09/xmldsig#' : 'xmldsig'
}

primitive_type_mappings = {
    "token"  : { "type" : "string"},
    "string" : { "type" : "string"},
    "nonNegativeInteger" : { "type" : "integer", "minimum" : 0},
    "anyType" : {}
}

def process_qname(qname):
    namespace = qname[1:qname.find('}')]
    token = qname[qname.find('}')+1:]
    return token, namespace

def get_type_from_elem(elem: XsdElement, local_id):
    t = elem.type
    ns = t.name[1:t.name.find('}')]
    if (ns == "http://www.w3.org/2001/XMLSchema"):
        # this should be an XSD primitive type
        mapped_type = primitive_type_mappings[t.local_name]
        return mapped_type
    else:
        # this needs to be a type in a namespace we know about
        mapped_ns = ns_mappings[ns]
        if (mapped_ns == local_id):
            return { "$ref" : f"#/$defs/{elem.type.local_name}"}
        else:
            return {"$ref" : f"{mapped_ns}.json#/$defs/{elem.type.local_name}"}

def get_ref(token, ns, context):
    mapped_ns = ns_mappings[ns]    
    if (mapped_ns == context):
        return f"#/$defs/{token}"
    else:
        return f"{mapped_ns}.json#/$defs/{token}"


for elementName, element in coreSchema.elements.items():
    print (f"Root element: {elementName} = {element}")
    token, ns = process_qname(element.type.name)
    js["oneOf"].append({
        '$ref' : get_ref(token, ns, 'core')
    })


def process_choice(choice: XsdGroup):
    if choice.model != 'choice':
        raise Exception(f"Wrong group type: {c.model}")
    oneOf = []
    for c in choice.iter_model():
        if not (type(c) is XsdElement):
            raise Exception (f"Non-element {c} encountered in choice {choice}")
        t = get_type_from_elem(c, 'core')
        oneOf.append({
            "type" : "object",
            "properties" : {
                c.local_name : t
            },
            "required" : [c.local_name]
        })
    return oneOf        
        

def process_atomic_restriction (sctype):
    r = {}
    print (f"  Atomic restriction: {sctype}")
    token, ns = process_qname(sctype.name)
    print (f"  [{ns}] {token}")
    print (f"  Root type {sctype.root_type}")
    print (f"  Base type {sctype.base_type}")
    print ("   Facets....")
    for k, f in sctype.facets.items():
        print (f"    facet {k} = {f}")
    print ("   Validators....")
    for v in list(sctype.validators):
        print (f"    validator {v}")
    mappedType = primitive_type_mappings.get(sctype.base_type.name)
    if mappedType:
        r = mappedType
        print (f"  Found primitive mapping to {mappedType}")
        if len(sctype.facets) > 0:
            raise Exception("Too many facets in primitive type")
        firstKey = sctype.facets.keys[0]
        if (firstKey != '{http://www.w3.org/2001/XMLSchema}pattern'):
            raise Exception (f"Unhandled facet {firstKey}")
        pattern = sctype.facets[firstKey]
        r['pattern'] = pattern.pattern
        if len(sctype.validators) > 0:
            raise Exception("Too many validators in primitive type")
        return token, r
    else:
        base_token, base_ns = process_qname(sctype.base_type.name)
        return token, { '$ref' : get_ref(base_token, base_ns, 'core')}


for typeName, sctype in coreSchema.types.items():
    print (f"Type: {typeName} = {type} ({type(sctype)}")
    if type(sctype) is XsdAtomicRestriction:
        token, typeDef = process_atomic_restriction(sctype)
        js["$defs"][token] = typeDef
        #print (dir(sctype))
    if type(sctype) is XsdComplexType:
        content = sctype.content
        print(f"  Content {content}")
        d = {}
        if type(content) is XsdGroup:
            if content.model == "sequence":
                print ("  SEQUENCE!")
                d = {
                    'type' : 'object',
                    'properties' : {},
                    'required' : []
                }
                innerChoice = None
                for c in list(content.iter_model()):
                    print (f"    > {c}")
                    if type(c) is XsdElement:
                        d['properties'][c.local_name] = get_type_from_elem(c, 'core')
                        if c.effective_min_occurs == 1:
                            d['required'].append(c.local_name)
                    elif type(c) is XsdGroup:
                        if innerChoice:
                            raise Exception ("Second group encountered")
                        if c.model != "choice":
                            raise Exception (f"Don't know what to do with inner {c}")
                        innerChoice = process_choice(c)
                    else:
                        raise Exception(f"Unknown element type {c}")
                if (innerChoice):
                    js["$defs"][sctype.local_name] = { 
                        'allOf' : [
                            d,
                            {'oneOf' : innerChoice}
                        ]
                    }
                else:
                    js["$defs"][sctype.local_name] = d
            elif content.model =="choice":
                print ("  CHOICE!")
                oneOf = process_choice(content)
                js["$defs"][sctype.local_name] = {'oneOf' : oneOf}
            
        else:
            raise Exception(f"Unknown content type {content}")


pprint(js)
Path('./core.json').write_text(json.dumps(js, indent=2))